/*****************************************************************************
 * grab-v4l2.c: v4l2 interface
 *****************************************************************************
 * $Id: grab-v4l2.c,v 1.21 2004/06/12 14:45:33 alainjj Exp $
 *****************************************************************************
 * Copyright (C) 2003 Alainjj
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************
 *
 * for the moment just the skeleton
 * very inspired from
 * xawtv/libng/plugins/drv0-v4l2.c
 *
*****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "config.h"
#include "strtab.h"
#include "grab.h"
#ifndef HAVE_V4L
struct GRABBER grab_v4l2; /* for compatibility... */
#else
#include "videodev.h"
#include "videodev2.h"
#include "colorspace.h"
#include "memcpy.h"
#include "blackborder.h"



/* prototypes */
static int grab_open (struct device_t *device, int sw, int sh,
                      int format, int pixmap, void *base, int width);
static int grab_close ();
static int grab_overlay (int x, int y, int width, int height, int format,
                         struct OVERLAY_CLIP *oc, int count);
static void *grab_scr (void *dest, int width, int height, int single,
                       int dest_xsize);
static void *grab_one (int format, int width, int height);
static int grab_tune (unsigned long freq);
static int grab_tuned ();
static int grab_input (int input, int norm);
static int grab_picture (int color, int bright, int hue, int contrast);
static int grab_audio (int mute, int volume, int *mode);
static int is525(int norm);
static int issecam(int norm);
static int stop_capture(void);
static int start_capture(void);
int changeformat(struct v4l2_format *);
static void* get_buf(int i);

/* pass 0/1 by reference */
static int one = 1, zero = 0;
static int grab = 0;
int overlay = 0;
static struct v4l2_framebuffer ov_fbuf;
static struct v4l2_clip ov_clips[32];
static int ov_error = 0;

static struct STRTAB device_cap[] = {
  {V4L2_CAP_VIDEO_CAPTURE, "capture"},
  {V4L2_CAP_VIDEO_OUTPUT,"output"},
  {V4L2_CAP_VIDEO_OVERLAY,"overlay"},
  {V4L2_CAP_VBI_CAPTURE,"vbi_capture"},
  {V4L2_CAP_RDS_CAPTURE,"rds"},
  {V4L2_CAP_TUNER,"tuner"},
  {V4L2_CAP_AUDIO,"audio"},
  {V4L2_CAP_READWRITE,"readwrite"},
  {V4L2_CAP_ASYNCIO,"asynci/o"},
  {V4L2_CAP_STREAMING,"streaming"},
  {-1,NULL}
};

#define MAX_INPUT   16
#define MAX_NORM    16
#define MAX_FORMAT  32

static struct {
  int  ninputs,nstds,nfmts;
  struct v4l2_capability	cap;
  struct v4l2_streamparm	streamparm;
  struct v4l2_input		inp[MAX_INPUT];
  struct v4l2_standard      	std[MAX_NORM];
  struct v4l2_fmtdesc		fmt[MAX_FORMAT];
  struct v4l2_queryctrl	brightness,contrast,hue,saturation;
  struct v4l2_queryctrl	mute,volume;
} drv;

struct GRABBER grab_v4l2 = {
  "v4l2",
  VIDEO_RGB16 | VIDEO_RGB24 | VIDEO_RGB32,
  0,
  NULL, NULL,
  grab_open,
  grab_close,
  grab_overlay,
  grab_scr,
  grab_one,
  grab_tune,
  grab_tuned,
  grab_input,
  grab_picture,
  grab_audio,
  is525,
  issecam,
  get_buf
};

extern int x11_bigendian;
extern struct GRABBER grab_dummy;
extern int bttvfd;
/* not static for xawdecode_scantv...*/
extern int vbifd;
extern void video_overlay
(int (*set_overlay)(int,int,int,int,int,struct OVERLAY_CLIP *,int));

/* screen grab */
struct v4l2_buffer *gb2;
static int gb_grab=0,gb_sync=0;
static int imgq=0,imgw=0;
extern int img;
static struct v4l2_requestbuffers reqbufs;

//struct video_mbuf gb_buffers = { 2 * 0x151000, 0, {0, 0x151000} };
static char **maps = NULL;

extern int nbufs;
struct v4l2_format gbformat2;
static double dur_frame = 0.04;
extern int have_xv;

static int
grab_open(struct device_t *device, int sw, int sh,
          int format, int pixmap, void *base, int bpl)
{
  int i;
  if (-1 != bttvfd)
    goto err;

  if (debug)
    fprintf (stderr, "v4l2: open\n");
  
  if (-1 == (bttvfd = open (device->video ? device->video : "/dev/video", O_RDWR)))
    {
      fprintf (stderr, "open %s: %s\n",
               device->video ? device->video : "/dev/video", strerror (errno));
      goto err;
    }
  
  if (-1 == ioctl(bttvfd,VIDIOC_QUERYCAP,&drv.cap)) 
    {
      fprintf (stderr, "driver is not v4l2\n");
      goto err;
    }

if ((drv.cap.capabilities & V4L2_CAP_VBI_CAPTURE) && (device->vbi))
    if (-1 == (vbifd = open (device->vbi, O_RDONLY)))
      {
	fprintf (stderr, "v4l2: cant't open %s for reading, %s\n",
		 device->vbi, strerror (errno));
	//goto err;
      }

  if (debug)
    {
      fprintf (stderr, "v4l2: driver=%.16s card=%.32s\n", 
	       drv.cap.driver,drv.cap.card);
      fprintf(stderr,"v4l2: capabilities =");
      for (i = 0; device_cap[i].str != NULL; i++)
        if (drv.cap.capabilities & device_cap[i].nr)
          fprintf (stderr, " %s", device_cap[i].str);
      fprintf (stderr, "\n");
    }
  
  
  for (drv.ninputs = 0; drv.ninputs < MAX_INPUT; drv.ninputs++) {
    drv.inp[drv.ninputs].index = drv.ninputs;
    if (-1 == ioctl(bttvfd, VIDIOC_ENUMINPUT, &drv.inp[drv.ninputs]))
      break;
  }
  for (drv.nstds = 0; drv.nstds < MAX_NORM; drv.nstds++) {
    drv.std[drv.nstds].index = drv.nstds;
    if (-1 == ioctl(bttvfd, VIDIOC_ENUMSTD, &drv.std[drv.nstds]))
      break;
  }
  for (drv.nfmts = 0; drv.nfmts < MAX_FORMAT; drv.nfmts++) {
    drv.fmt[drv.nfmts].index = drv.nfmts;
    drv.fmt[drv.nfmts].type  = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl(bttvfd, VIDIOC_ENUM_FMT, &drv.fmt[drv.nfmts]))
      break;
  }
  
  drv.streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  ioctl(bttvfd,VIDIOC_G_PARM,&drv.streamparm);
  
  drv.brightness.id = V4L2_CID_BRIGHTNESS;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.brightness))
    drv.brightness.id = 0;
  drv.contrast.id = V4L2_CID_CONTRAST;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.contrast))
    drv.contrast.id = 0;
  drv.hue.id = V4L2_CID_HUE;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.hue))
    drv.hue.id = 0;
  drv.saturation.id = V4L2_CID_SATURATION;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.saturation))
    drv.saturation.id = 0;
  drv.mute.id = V4L2_CID_AUDIO_MUTE;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.mute))
    drv.mute.id = 0;
  drv.volume.id = V4L2_CID_AUDIO_VOLUME;
  if(-1 == ioctl(bttvfd,VIDIOC_QUERYCTRL,&drv.volume) ||
     drv.volume.minimum==drv.volume.maximum)
    drv.volume.id = 0;
  //exit(0);

  grab_v4l2.norms=malloc(sizeof(struct STRTAB) * (drv.nstds+1));
  for (i = 0; i < drv.nstds; i++) {
    grab_v4l2.norms[i].nr  = i;
    grab_v4l2.norms[i].str = drv.std[i].name;
  }
  grab_v4l2.norms[i].nr  = -1;
  grab_v4l2.norms[i].str = NULL;
  if(debug) {
    printf("norms=");
    for(i=0;i<drv.nstds;i++) printf(" %s",grab_v4l2.norms[i].str);
    printf("\n");
  }

  grab_v4l2.inputs=malloc(sizeof(struct STRTAB) * (drv.ninputs+1));
  for (i = 0; i < drv.ninputs; i++) {
    grab_v4l2.inputs[i].nr  = i;
    grab_v4l2.inputs[i].str = drv.inp[i].name;
  }
  grab_v4l2.inputs[i].nr  = -1;
  grab_v4l2.inputs[i].str = NULL;
  if(debug) {
    printf("inputs=");
    for(i=0;i<drv.ninputs;i++) printf(" %s",grab_v4l2.inputs[i].str);
    printf("\n");
  }

  if(pixmap==0) return 0; // for xawdecode_scantv

  /**************************************************/
  gbformat2.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  gbformat2.fmt.pix.width = sw;
  gbformat2.fmt.pix.height = sh;
  gbformat2.fmt.pix.field = V4L2_FIELD_ANY;
  switch (pixmap)
    {
    case VIDEO_GRAY:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_GREY\n");
      gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_GREY;
      break;
    case VIDEO_RGB08:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_HI240\n");
      gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_HI240;
      break;
    case VIDEO_RGB15:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_RGB555\n");
      if(x11_bigendian)
	gbformat2.fmt.pix.pixelformat  = V4L2_PIX_FMT_RGB555X;
      else
	gbformat2.fmt.pix.pixelformat  = V4L2_PIX_FMT_RGB555;
      break;
    case VIDEO_RGB16:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_RGB565\n");
      if(x11_bigendian)
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565X;
      else
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;
      break;
    case VIDEO_RGB24:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_RGB24\n");
      if(x11_bigendian)
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
      else
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      break;
    case VIDEO_RGB32:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_BGR32 /32)\n");
      if(x11_bigendian)
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB32;
      else
	gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR32;
      break;
    case VIDEO_YUY2:
      if (debug)
        fprintf(stderr, "grab_open: setting format to V4L2_PIX_FMT_YUYV\n");
      gbformat2.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
      break;
    default:
      fprintf(stderr, "format not implemented\n");
      goto err;
    }
  if(-1 == ioctl(bttvfd,VIDIOC_S_FMT, &gbformat2)) {
    perror("ioctl VIDIOC_S_FMT");
    goto err;
  }
  
  /* frame buffer */
  if (-1 == ioctl (bttvfd, VIDIOC_G_FBUF, &ov_fbuf)) {
    perror ("ioctl VIDIOC_G_FBUF");
    fprintf(stderr, "classical overlay is disabled");
    ov_error=1;
  } else {
    if (debug)
      fprintf (stderr, "  fbuffer : size=%dx%d bpl=%d\n",
	       ov_fbuf.fmt.width, ov_fbuf.fmt.height,
	       ov_fbuf.fmt.bytesperline);
    if(ov_fbuf.base==NULL) { 
      fprintf(stderr,
	      "WARNING: video memory base unknown, may be caused by a problem\n"
	      "  with xawdecode_v4l-conf or a non-availability of DGA\n"
	      "  and frame buffer devices: CLASSICAL OVERLAY IS DISABLED !\n");
      ov_error=1;
    }
  }
  
  start_capture();
  
  return 0;
  
 err:
  if (bttvfd != -1) {
      close (bttvfd);
      bttvfd = -1;
  }
  if (vbifd != -1)  {
    close (vbifd);
    vbifd = -1;
  }
  if(gb2!=NULL) free(gb2);
  if(maps!=NULL) free(maps);
  return -1;
}


static int
grab_close ()
{
  if (-1 == bttvfd)
    return 0;
  grab_audio (1, -1, NULL);
  if(grab) stop_capture();
  close(bttvfd);
  bttvfd = -1;
  close(vbifd);
  vbifd = -1;
  free(grab_v4l2.norms); grab_v4l2.norms = NULL;
  free(grab_v4l2.inputs); grab_v4l2.inputs = NULL;
  return 0;
}


static int
grab_overlay (int x, int y, int width, int height, int format,
              struct OVERLAY_CLIP *oc, int count)
{
  struct v4l2_format win;
  int i;

  if(bttvfd == -1) return -1;
  if(ov_error) { overlay = 0;  return 0; }

  if(grab) stop_capture();
  if (width == 0 || height == 0)
    {
      if (debug)
        fprintf (stderr, "v4l2: overlay off\n");
      if( -1 == ioctl (bttvfd, VIDIOC_OVERLAY, &zero))
	perror("ioctl VIDIOC_OVERLAY");
      overlay = 0;
      changeformat(&gbformat2);
      return 0;
    }
  memset(&win,0,sizeof(win));
  win.type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
  win.fmt.win.w.left    = x;
  win.fmt.win.w.top     = y;
  win.fmt.win.w.width   = width;
  win.fmt.win.w.height  = height;
  
  if (drv.cap.capabilities & V4L2_FBUF_CAP_LIST_CLIPPING) {
    win.fmt.win.clips      = ov_clips;
    win.fmt.win.clipcount  = count;
    
    for (i = 0; i < count; i++) {
      ov_clips[i].next = (i+1 == count) ? NULL : &ov_clips[i+1];
      ov_clips[i].c.left = oc[i].x1;
      ov_clips[i].c.top = oc[i].y1;
      ov_clips[i].c.width = oc[i].x2 - oc[i].x1 /* XXX */ ;
      ov_clips[i].c.height = oc[i].y2 - oc[i].y1;
      if (debug)
	fprintf (stderr, "v4l2: clip=%dx%d+%d+%d\n",
		 ov_clips[i].c.width, ov_clips[i].c.height,
		 ov_clips[i].c.left, ov_clips[i].c.top);
	}
  }
  
  if(-1 == ioctl(bttvfd,VIDIOC_S_FMT, &win)) {
    perror("ioctl VIDIOC_S_FMT");
    return -1;
  }
  
  if(!overlay) {
    if( -1 == ioctl (bttvfd, VIDIOC_OVERLAY, &one)) {
      perror("ioctl VIDIOC_OVERLAY");
      return -1;
    }
    overlay = 1;
  }

  return 0;
}


static int
grab_queue (struct v4l2_buffer *gb)
{
  if (debug > 1)
    fprintf (stderr, "g%d", gb->index);

  if(overlay) {fprintf(stderr,"BIG PB\n");exit(1);}
  if (-1 == ioctl (bttvfd, VIDIOC_QBUF, gb))
    {
      if (errno == EAGAIN)
        fprintf (stderr, "grabber chip can't sync (no station tuned in?)\n");
      else
        fprintf (stderr, "ioctl VIDIOC_QBUF(%d): %s\n",
                 gb->index, strerror (errno));
      return -1;
    }

  gb_grab++;
  if (debug > 1)
    fprintf (stderr, "* ");

  return 0;
}

static int
grab_wait (struct v4l2_buffer  *gb)
{
  fd_set  rdset;
  struct timeval tv;
  int r;

  if (debug)
    fprintf (stderr, "s%d", gb->index);

  do {
    tv.tv_sec  = 5;
    tv.tv_usec = 0;
    FD_ZERO(&rdset);
    FD_SET(bttvfd, &rdset);
    r = select(bttvfd + 1, &rdset, NULL, NULL, &tv);
  } while(r==-1 && errno == EINTR);
  if(r==-1) {
    perror("grab_wait: select");
    return -1;
  }
  if(r==0) {
    fprintf(stderr,"grab_wait: select timeout\n");
    return -1;
  }

  if (-1 == ioctl (bttvfd, VIDIOC_DQBUF, gb))
    {
      perror ("ioctl VIDIOC_DQBUF");
      return -1;
    }
  
  gb_sync++;
  if (debug)
    fprintf (stderr, "* ");
  
  return 0;
}

static inline void reset(void) {
  fprintf(stderr,"total reset\n");
  for (imgw=0;imgw<nbufs;imgw++) grab_wait(&gb2[imgw]);
  gb_grab=gb_sync=imgq=imgw=0;
}


/* with v4l2 the buffers are not associated to a format....*/
char * get_img2(void) {
  if(imgq!=gb_grab%nbufs || imgw!=gb_sync%nbufs)
    fprintf(stderr,"PB imgq=%d gb_grab=%d imgw=%d gb_sync=%d nbufs=%d\n",
	    imgq,gb_grab,imgw,gb_sync,nbufs);
  
  while ((gb_grab-gb_sync)<nbufs) {
    if (-1 == grab_queue(&gb2[imgq]))  {
      reset();
      return NULL;
    }
    imgq++;
    if (imgq>=nbufs) imgq=0;
  }
  
  if (-1 == grab_wait(&gb2[imgw])) {
    reset();
    return NULL;
  }
  
  img=imgw;
  
  imgw++;
  if (imgw==nbufs) imgw=0;
  videostampmin = gb2[img].timestamp.tv_usec/1000000.0+
    gb2[img].timestamp.tv_sec + dur_frame;
  videostampmax = videostampmin + dur_frame;
  if(debug>=3) {
    double now;
    struct timeval tv;
    gettimeofday (&tv, NULL);
    now = tv.tv_usec/1000000.0+tv.tv_sec;
    printf("img=%d min=%.4lf max=%.4lf now=%.4lf\n",img,
	   videostampmin-(int)videostampmin,
	   videostampmax-(int)videostampmax,now-(int)now);
  }
  return maps[img];
}


static int stop_capture(void) {
  int i,type;
  while (gb_grab!=gb_sync) grab_wait(&gb2[imgw]);
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if(-1 == ioctl(bttvfd,VIDIOC_STREAMOFF,&type)) {
    perror("ioctl VIDIOC_STREAMOFF");
    return -1;
  }
  for(i = 0; i < nbufs; i++)
    if(maps[i]!=NULL) munmap(maps[i],gb2[i].length);
  free(maps);free(gb2);
  maps=NULL; gb2=NULL;
  grab = 0;
  return 0;
}

static int start_capture(void) {
  int i,type;
  if(overlay) video_overlay (NULL);;
  gb_grab=gb_sync=imgq=imgw=0;
  reqbufs.count  = nbufs_default;
  reqbufs.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  reqbufs.memory = V4L2_MEMORY_MMAP;
  if (-1 == ioctl(bttvfd, VIDIOC_REQBUFS, &reqbufs)) {
    perror("ioctl VIDIOC_REQBUFS");
    return -1;
  }
  nbufs = reqbufs.count;
  if(debug) 
    fprintf(stderr,"allocated %d kernel buffers\n",nbufs);
  gb2=(struct v4l2_buffer *) malloc(nbufs*sizeof(struct v4l2_buffer));
  maps=(char **) malloc(nbufs*sizeof(char*));
  memset(maps, 0, nbufs * sizeof(char *));
  for (i = 0; i < nbufs; i++) {
    gb2[i].index  = i;
    gb2[i].type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    gb2[i].memory = V4L2_MEMORY_MMAP;
    if (-1 == ioctl(bttvfd, VIDIOC_QUERYBUF, &gb2[i], 0)) {
      perror("ioctl VIDIOC_QUERYBUF");
      return -1;
    }
    maps[i] = mmap(NULL, gb2[i].length, PROT_READ | PROT_WRITE, MAP_SHARED,
		   bttvfd, gb2[i].m.offset);
    if(maps[i] == MAP_FAILED) {
      perror("mmap");
      return -1;
    }
  }
  if(-1 == ioctl(bttvfd,VIDIOC_STREAMON,&type)) {
    perror("ioctl VIDIOC_STREAMON");
    return -1;
  }
  grab = 1;
  return 0;
}


int changeformat(struct v4l2_format * fmt) {
  if(grab)
    if(-1 == stop_capture()) return -1;

  if(-1 == ioctl(bttvfd,VIDIOC_S_FMT, fmt)) {
    perror("ioctl VIDIOC_S_FMT");
    return -1;
  }
  if(!overlay)
    if(-1 == start_capture()) return -1;
    
  return 0;
}


static void *
grab_scr (void *dest, int width, int height, int single, int dest_xsize)
{
  void *buf;
  int yb;
  if(gbformat2.fmt.pix.width!=width || gbformat2.fmt.pix.height!=height) {
    gbformat2.fmt.pix.width=width;
    gbformat2.fmt.pix.height=height;
    if(changeformat(&gbformat2) == -1) {
      return 0;
    }
  }
  buf = get_img2();
  if(buf == NULL) return NULL;
  //rgb24_to_rgb32 (dest, buf, width, height);
  yb = get_ybar(height);
  if(yb>0 && have_xv) yb--;
  yb *= gbformat2.fmt.pix.bytesperline;
  if(width==dest_xsize)
    fast_memcpy (dest+yb, buf+yb, height*gbformat2.fmt.pix.bytesperline-2*yb);
  else {
    /* only for directvideo */
    int bpp=gbformat2.fmt.pix.bytesperline/gbformat2.fmt.pix.width;
    int y;
    for (y=0; y<height; y++)
      fast_memcpy(dest+y*dest_xsize*bpp, buf+y*width*bpp, width*bpp);
  }
  return dest;
}

static void *
grab_one (int x11_format, int width, int height)
{
  struct v4l2_format gbformat2_one = gbformat2;
  static void *dest=NULL;
  void *buf;
  int bpp;
  int overlay2;

  gbformat2_one.fmt.pix.width = width;
  gbformat2_one.fmt.pix.height = height;
  switch(x11_format)
    {
    case VIDEO_GRAY:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_GREY\n");
      gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_GREY;
      bpp=1;
      break;
    case VIDEO_RGB08:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_HI240\n");
      gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_HI240;
      bpp=1;
      break;
    case VIDEO_RGB15:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_RGB555\n");
      if(x11_bigendian)
	gbformat2_one.fmt.pix.pixelformat  = V4L2_PIX_FMT_RGB555X;
      else
	gbformat2_one.fmt.pix.pixelformat  = V4L2_PIX_FMT_RGB555;
      bpp=2;
      break;
    case VIDEO_RGB16:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_RGB565\n");
      if(x11_bigendian)
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565X;
      else
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB565;
      bpp=2;
      break;
    case VIDEO_RGB24:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_RGB24\n");
      if(x11_bigendian)
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
      else
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      bpp=3;
      break;
    case VIDEO_RGB32:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_RGB24\n");
      if(x11_bigendian)
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB32;
      else
	gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR32;	
      bpp=4;
      break;
    case VIDEO_YUY2:
      if (debug)
        fprintf(stderr, "grab_one: setting format to V4L2_PIX_FMT_YUYV\n");
      gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
      bpp=2;
      break;
    default:
      if (debug)
        fprintf(stderr, "grab_one: format = UNKNOWN / VIDEO_PALETTE_RGB24\n");
      gbformat2_one.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
      bpp=3;
      break;
    }
  overlay2 = overlay;
  if(overlay2) video_overlay (NULL);
  changeformat(&gbformat2_one);
  buf = get_img2();
  if(buf!=NULL) {
    dest = realloc(dest,width*height*bpp);
    fast_memcpy(dest, buf, width*height*bpp);
  }
  // must be copied before changeformat changes all the mapped...
  changeformat(&gbformat2);
  if(overlay2) video_overlay (grab_overlay);
  if(buf==NULL) return NULL;
  return dest;
}

/* ---------------------------------------------------------------------- */

static int
grab_tune (unsigned long freq)
{
  struct v4l2_frequency f;
  memset(&f,0,sizeof(f));
  f.type = V4L2_TUNER_ANALOG_TV;
  f.frequency = freq;
  if(ioctl(bttvfd, VIDIOC_S_FREQUENCY, &f) == -1)
    perror("VIDIOC_S_FREQUENCY");
  return 0;
}

static int
grab_tuned ()
{
  struct v4l2_tuner tuner;
  
  memset(&tuner,0,sizeof(tuner));
  if (-1 == ioctl(bttvfd,VIDIOC_G_TUNER,&tuner))
    return 0;
  return tuner.signal ? 1 : 0;
}

static int
grab_input (int input, int norm)
{
  static int prev_norm = -1;
  if (-1 != input) {
    if (debug)
      fprintf (stderr, "v4l2: input: %d\n", input);
    if( -1 == ioctl (bttvfd, VIDIOC_S_INPUT, &drv.inp[input].index))
      perror("ioctl VIDIOC_S_INPUT");
  }
  if (-1 != norm)
    {
      if (debug)
        fprintf (stderr, "v4l2: norm : %d\n", norm);
      if(norm != prev_norm) {
	int overlay2 = overlay;
	if(overlay2) video_overlay (NULL);
	if(grab) stop_capture();
	if( -1 == ioctl(bttvfd, VIDIOC_S_STD, &drv.std[norm].id)) {
	  perror("ioctl VIDIOC_S_STD");
	  return -1;
	}
	if(overlay2)
	  video_overlay (grab_overlay);
	else
	  start_capture();
	prev_norm = norm;
	if(is525(norm)) 
	  dur_frame = 10001.0/30000;
	else
	  dur_frame = 1.0/25;
      }
    }
  return 0;
}

static inline int control(int x,struct v4l2_queryctrl *ctrl) {
  struct v4l2_control c;
  if(x == -1) return 0;
  if(ctrl->id == 0) return -1;
  c.id = ctrl->id;
  c.value = (float)x/65535*(ctrl->maximum-ctrl->minimum)+ctrl->minimum+0.5;
  if(-1 == ioctl(bttvfd, VIDIOC_S_CTRL, &c)) {
    perror("ioctl VIDIOC_S_CTRL");
    return -1;
  }
  return 0;
}
		   
static int
grab_picture (int color, int bright, int hue, int contrast)
{
  control(color,&drv.saturation);
  control(bright,&drv.brightness);
  control(hue,&drv.hue);
  control(contrast,&drv.contrast);
  return 0;
}

static int
grab_audio (int mute, int volume, int *mode)
{
  struct v4l2_control c;
  struct v4l2_tuner tuner;
  if(mute !=-1 && drv.mute.id) {
    c.id = V4L2_CID_AUDIO_MUTE;
    c.value = mute;
    if(-1 == ioctl(bttvfd, VIDIOC_S_CTRL,&c))
      perror("VIDIOC_S_CTRL, V4L2_CID_AUDIO_MUTE");
  }
  if(volume !=-1 && drv.volume.id) {
    c.id = V4L2_CID_AUDIO_VOLUME;
    c.value = volume;
    if(-1 == ioctl(bttvfd, VIDIOC_S_CTRL,&c))
      perror("VIDIOC_S_CTRL, V4L2_CID_AUDIO_VOLUME");
  }
  if(mode) {
    memset(&tuner,0,sizeof(tuner));
    if(-1 == ioctl(bttvfd,VIDIOC_G_TUNER,&tuner))
      perror("VIDIOC_G_TUNER");
    if(*mode) {
      switch(*mode) {
      case 1: tuner.audmode = V4L2_TUNER_MODE_MONO; break;
      case 2: tuner.audmode = V4L2_TUNER_MODE_STEREO; break;
      }
      if(-1 == ioctl(bttvfd,VIDIOC_S_TUNER,&tuner))
	perror("VIDIOC_S_TUNER");
      if(-1 == ioctl(bttvfd,VIDIOC_G_TUNER,&tuner))
	perror("VIDIOC_G_TUNER");
    }
    switch(tuner.audmode) {
    case V4L2_TUNER_MODE_MONO: *mode=1; break;
    case V4L2_TUNER_MODE_STEREO: *mode=2; break;
    default: *mode=0;
    }
  }
  return 0;
}

static int is525(int norm) {
  return drv.std[norm].id & V4L2_STD_525_60;
}

static int issecam(int norm) {
  return drv.std[norm].id & V4L2_STD_SECAM;
}

static void* get_buf(int i){
  return maps[i];
}
#endif
