/*****************************************************************************
 * subtitles_ui.c: Enter your own subtitle page
 *****************************************************************************
 * $Id: subtitles_ui.c,v 1.13 2004/05/30 14:22:20 pingus77 Exp $
 *****************************************************************************
 * Copyright (C) 2004 Pingus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/AsciiText.h>

#ifdef HAVE_XPM
# include <X11/xpm.h>
#endif

#include "xawdecode.h"
#include "grab.h"
#include "x11.h"

#ifdef HAVE_XPM
# include "xawdecode.xpm"
#endif

# include "popup_ui.h"

/*-------------------------------------------------------------------------*/
extern void save_cb (Widget widget, XtPointer clientdata, XtPointer call_data);
extern void ToggleSubAction (Widget, XEvent *, String *, Cardinal *);
extern void OptionsAction (Widget, XEvent *, String *, Cardinal *);
extern void action_cb (Widget widget, XtPointer clientdata, XtPointer call_data);
extern Widget app_shell;
extern Display *dpy;
extern Atom wm_protocols[2];
extern int debug;
extern char *subtitles_pageno_str;
extern int subactive;
extern int submethod;
extern int xawpopup;

static Widget subpage_shell, subpage_paned;
static Widget t_subpagenum;
static Widget c_subpagenum;
static Widget c_subtitles;
static Widget c_openmain;

#ifdef HAVE_XPM
  Pixmap icon_pixmap;
  Pixmap icon_shapemask;
#endif

/*-------------------------------------------------------------------------*/

struct CALL_ACTION call_subtitles = { 0, "ToggleSub", {NULL} };
struct CALL_ACTION call_options_3 = { 0, "Options", {NULL} };

/*--------------------------------------------------------------------------*/

static void
close_subpagewin (Widget widget, XtPointer clientdata, XtPointer call_data)
{
  XtCallActionProc (subpage_shell, "SubPage", NULL, NULL, 0);
}

void
SubPageAction (Widget widget, XEvent * event,
              String * params, Cardinal * num_params)
{
  static int mapped = 0, first = 1;
  submethod = 1;
  if (debug)
  fprintf(stderr,"subtitles_pageno_str SubPageAction: %s\n",subtitles_pageno_str);

  if (event && event->type == ClientMessage)
    {
      if (event->xclient.data.l[0] == wm_protocols[1])
        {
          if (debug)
            fprintf (stderr, "Config: wm_save_yourself\n");
          XSetCommand (XtDisplay (subpage_shell), XtWindow (subpage_shell), NULL,
                       0);
          return;
        }
    }

  if (subpage_shell == NULL)
    {
      fprintf(stderr, "SubPageAction: ooops: subpage_shell is NULL\n");
            if (xawpopup)
              PopupMessage("SubPageAction:\nooops: subpage_shell is NULL\n");
      return;
    }

  if(subactive == 1)
    {
      fprintf(stderr, "SubPageAction: Can't add a specific subtitle page number if subtitle is active\n");
            if (xawpopup)
              PopupMessage("Can't add a specific subtitle\npage number if subtitle is active\n");
      return;
    }

  if (mapped)
    {
      XtPopdown (subpage_shell);
      mapped = 0;
    }
  else
    {
      XtPopup (subpage_shell, XtGrabNone);
      mapped = 1;
      if (first)
        {
          XSetWMProtocols (XtDisplay (subpage_shell), XtWindow (subpage_shell),
                           wm_protocols, 2);
          first = 0;
        }
    }
}

/*--------------------------------------------------------------------------*/

void
create_subpagewin()
{
  Widget c;

  subpage_shell = XtVaAppCreateShell ("SubPage", "Xawdecode",
                                   topLevelShellWidgetClass,
                                   dpy, XtNclientLeader, app_shell, XtNvisual,vinfo->visual,
				   XtNcolormap,colormap, XtNdepth,vinfo->depth,
				   NULL);

#ifdef HAVE_XPM
  XpmCreatePixmapFromData(XtDisplay(subpage_shell),
                          RootWindowOfScreen(XtScreen(subpage_shell)),
                          icon_xpm,
                          &icon_pixmap,
                          &icon_shapemask, NULL);
  XtVaSetValues(subpage_shell, XtNiconPixmap, icon_pixmap, NULL);
#endif

  subpage_paned =
    XtVaCreateManagedWidget ("paned", panedWidgetClass, subpage_shell, NULL);

  /* Subtitle page number */

  c_subpagenum= c = XtVaCreateManagedWidget ("lsubpagenum", labelWidgetClass, subpage_paned,
                           PANED_FIX, NULL);
  t_subpagenum = c = XtVaCreateManagedWidget ("subpagenum", asciiTextWidgetClass, subpage_paned,
                                                PANED_FIX, NULL);
  XtVaSetValues (t_subpagenum, XtNstring, subtitles_pageno_str);
  if (subtitles_pageno_str == NULL) subtitles_pageno_str = "all";

  
/* launch Subtitles Action */
  c_subtitles = c = XtVaCreateManagedWidget ("lsubtitles", commandWidgetClass, subpage_paned, PANED_FIX, NULL);
    XtAddCallback (c, XtNcallback, close_subpagewin, (XtPointer) NULL);   
    XtAddCallback (c, XtNcallback, action_cb, (XtPointer) & call_subtitles);   

    
/* Open Main Options Menu */
  c_openmain = c = XtVaCreateManagedWidget ("lopenmain", commandWidgetClass, subpage_paned, PANED_FIX, NULL);
    XtAddCallback (c, XtNcallback, action_cb, (XtPointer) & call_options_3);
          

/* Button to close Subtitle Page Menu */

  c = XtVaCreateManagedWidget ("lsubsave", commandWidgetClass, subpage_paned, PANED_FIX, NULL);
  if (debug)
    fprintf(stderr,"submethod: %d\n",submethod);
  XtAddCallback (c, XtNcallback, save_cb, (XtPointer) NULL);
  XtAddCallback (c, XtNcallback, close_subpagewin, (XtPointer) NULL);

}

void update_subpagenum(void)
{
  XtVaGetValues (t_subpagenum, XtNstring, &subtitles_pageno_str, NULL);
}
